/* Copyright © 2010 Brandon L Black <blblack@gmail.com>
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "dnsio_udp.h"

#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/uio.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <pthread.h>
#include <sys/mman.h>

#include "conf.h"
#include "dnswire.h"
#include "dnspacket.h"
#if DNSCURVE
#include "dnscurve.h"
#endif

#ifndef SOL_IPV6
#define SOL_IPV6 IPPROTO_IPV6
#endif

#ifndef SOL_IP
#define SOL_IP IPPROTO_IP
#endif

static void udp_sock_opts_v4(const int sock V_UNUSED, const bool any_addr) {
    const int opt_one V_UNUSED = 1;
    // If all variants we know of don't exist, we simply assume the IP
    //  stack will *not* set the DF bit on UDP packets.  We may need
    //  more variants here for other operating systems.
#if defined IP_MTU_DISCOVER && defined IP_PMTUDISC_DONT
    const int mtu_type = IP_PMTUDISC_DONT;
    if(setsockopt(sock, SOL_IP, IP_MTU_DISCOVER, &mtu_type, sizeof (mtu_type)) == -1)
        log_fatal("Failed to disable Path MTU Discovery for UDP socket: %s", logf_errno());
#elif defined IP_DONTFRAG
    const int opt_zero = 0;
    if(setsockopt(sock, SOL_IP, IP_DONTFRAG, &opt_zero, sizeof (opt_zero)) == -1)
        log_fatal("Failed to disable DF bit for UDP socket: %s", logf_errno());
#endif

    // This is just a latency hack, it's not necessary for correct operation
#if defined IP_TOS && defined IPTOS_LOWDELAY
    const int opt_tos = IPTOS_LOWDELAY;
    if(setsockopt(sock, SOL_IP, IP_TOS, &opt_tos, sizeof opt_tos) == -1)
        log_fatal("Failed to set IPTOS_LOWDELAY on UDP socket: %s", logf_errno());
#endif

    if(any_addr) {
#if HAVE_DECL_IP_PKTINFO
        if(setsockopt(sock, SOL_IP, IP_PKTINFO, &opt_one, sizeof opt_one) == -1)
            log_fatal("Failed to set IP_PKTINFO on UDP socket: %s", logf_errno());
#elif HAVE_DECL_IP_RECVDSTADDR && HAVE_DECL_IP_SENDSRCADDR
        // we don't use SENDSRCADDR directly, but it seems most smart implementors
        //  define it as an alias to RECVDSTADDR.  Importantly: MacOS, which does
        //  not implement the sending part of this magic, does not declare SENDSRCADDR
#  if IP_RECVDSTADDR != IP_SENDSRCADDR
#    error Your platform violates some gdnsd assumptions (IP_RECVDSTADDR != IP_SENDSRCADDR)
#  endif
        if(setsockopt(sock, SOL_IP, IP_RECVDSTADDR, &opt_one, sizeof opt_one) == -1)
            log_fatal("Failed to set IP_RECVDSTADDR on UDP socket: %s", logf_errno());
#else
        log_fatal("IPv4 any-address '0.0.0.0' not supported for DNS listening on your platform (no IP_PKTINFO or IP_RECVDSTADDR+IP_SENDSRCADDR)");
#endif
    }
}

/* Here, we assume that if neither IPV6_USE_MIN_MTU or IPV6_MTU is
 *  available that the kernel will fragment for us by default.  This
 *  may or may not be a safe assumption on all OS's.
 * To test: set up an environment where one link in the client<->server
 *  path has a smaller MTU than the server interface MTU, and the
 *  server's interface MTU is >1280. Send an IPv6 query that results in
 *  a response greater than the path MTU, but smaller than the server's
 *  interface MTU.
 *  If the response does not reach the client, this platform is broken,
 *  and we need to find a platform-specific way to make it fragment to
 *  1280 or disable IPv6 completely for this platform.
 */

static void udp_sock_opts_v6(const int sock, const bool any_addr) {
    const int opt_one = 1;

#if defined IPV6_USE_MIN_MTU
    if(setsockopt(sock, SOL_IPV6, IPV6_USE_MIN_MTU, &opt_one, sizeof opt_one) == -1)
        log_fatal("Failed to set IPV6_USE_MIN_MTU on UDP socket: %s", logf_errno());
#elif defined IPV6_MTU
#    ifndef IPV6_MIN_MTU
#      define IPV6_MIN_MTU 1280
#    endif
    const int min_mtu = IPV6_MIN_MTU;
    if(setsockopt(sock, SOL_IPV6, IPV6_MTU, &min_mtu, sizeof min_mtu) == -1)
        log_fatal("Failed to set IPV6_MTU on UDP socket: %s", logf_errno());
#endif

    if(setsockopt(sock, SOL_IPV6, IPV6_V6ONLY, &opt_one, sizeof opt_one) == -1)
        log_fatal("Failed to set IPV6_V6ONLY on UDP socket: %s", logf_errno());

#if defined IPV6_TCLASS && defined IPTOS_LOWDELAY
    const int opt_tos = IPTOS_LOWDELAY;
    if(setsockopt(sock, SOL_IPV6, IPV6_TCLASS, &opt_tos, sizeof opt_tos) == -1)
        log_fatal("Failed to set IPTOS_LOWDELAY on UDP socket: %s", logf_errno());
#endif

    if(any_addr) {
#if HAVE_DECL_IPV6_RECVPKTINFO
        if(setsockopt(sock, SOL_IPV6, IPV6_RECVPKTINFO, &opt_one, sizeof opt_one) == -1)
            log_fatal("Failed to set IPV6_RECVPKTINFO on UDP socket: %s", logf_errno());
#elif HAVE_DECL_IPV6_PKTINFO
        if(setsockopt(sock, SOL_IPV6, IPV6_PKTINFO, &opt_one, sizeof opt_one) == -1)
            log_fatal("Failed to set IPV6_PKTINFO on UDP socket: %s", logf_errno());
#else
        log_fatal("IPv6 any-address '::' not supported for DNS listening on your platform (no IPV6_RECVPKTINFO/IPV6_PKTINFO)");
#endif
    }
}

bool udp_sock_setup(dns_addr_t *addrconf) {
    dmn_assert(addrconf);

    const anysin_t* asin = &addrconf->addr;

    const bool isv6 = asin->sa.sa_family == AF_INET6 ? true : false;
    dmn_assert(isv6 || asin->sa.sa_family == AF_INET);

    const int sock = socket(isv6 ? PF_INET6 : PF_INET, SOCK_DGRAM, gdnsd_getproto_udp());
    if(sock == -1) log_fatal("Failed to create IPv%c UDP socket: %s", isv6 ? '6' : '4', logf_errno());

    const int opt_one = 1;
    if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt_one, sizeof opt_one) == -1)
        log_fatal("Failed to set SO_REUSEADDR on UDP socket: %s", logf_errno());

    int opt_size;
    socklen_t size_size = sizeof(opt_size);

    if(addrconf->udp_rcvbuf) {
        opt_size = addrconf->udp_rcvbuf;
        if(setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &opt_size, sizeof(opt_size)) == -1)
            log_fatal("Failed to set SO_RCVBUF to %u for UDP socket %s: %s", opt_size,
                logf_anysin(asin), logf_errno());
    }
    else {
        // Enforce a basic minimum SO_RCVBUF of (8 * DNS_RECV_SIZE) (32K)
        if(getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &opt_size, &size_size) == -1)
            log_fatal("Failed to get SO_RCVBUF on UDP socket: %s", logf_errno());

        if(opt_size < (int)(8 * DNS_RECV_SIZE)) {
            opt_size = 8 * DNS_RECV_SIZE;
            if(setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &opt_size, sizeof(opt_size)) == -1)
                log_fatal("Failed to set SO_RCVBUF to %u for UDP socket %s: %s", opt_size,
                    logf_anysin(asin), logf_errno());
        }
    }

    if(addrconf->udp_sndbuf) {
        opt_size = addrconf->udp_sndbuf;
        if(setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &opt_size, sizeof(opt_size)) == -1)
            log_fatal("Failed to set SO_SNDBUF to %u for UDP socket %s: %s", opt_size,
                logf_anysin(asin), logf_errno());
    }
    else {
        // Enforce a basic minimum SO_SNDBUF to the lesser of (4 * outer_max_response)
        //  (typically ~16K + curve overhead) or 128K.
        if(getsockopt(sock, SOL_SOCKET, SO_SNDBUF, &opt_size, &size_size) == -1)
            log_fatal("Failed to get SO_SNDBUF on UDP socket: %s", logf_errno());

        int desired_sndbuf = 4 * gconfig.outer_max_response;
        if(desired_sndbuf > 131072)
            desired_sndbuf = 131072;
        if(opt_size < desired_sndbuf) {
            opt_size = desired_sndbuf;
            if(setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &opt_size, sizeof(opt_size)) == -1)
                log_fatal("Failed to set SO_SNDBUF to %u for UDP socket %s: %s", opt_size,
                    logf_anysin(asin), logf_errno());
        }
    }


    if(isv6)
        udp_sock_opts_v6(sock, is_any_addr(asin));
    else
        udp_sock_opts_v4(sock, is_any_addr(asin));

    addrconf->udp_sock = sock;

    if(bind(sock, (const struct sockaddr*)asin, isv6 ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in))) {
        if(addrconf->late_bind_secs && errno == EADDRNOTAVAIL) {
            addrconf->udp_need_late_bind = 1;
            log_info("UDP DNS socket %s not yet available, will attempt late bind every %u seconds", logf_anysin(asin), addrconf->late_bind_secs);
            return ntohs(isv6 ? asin->sin6.sin6_port : asin->sin.sin_port) < 1024 ? true : false;
        }
        log_fatal("Failed to bind() UDP socket to %s: %s", logf_anysin(asin), logf_errno());
    }

    return false;
}

#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS MAP_ANON
#endif

// A reasonable guess for v4/v6 dstaddr pktinfo + cmsg header?
#define CMSG_BUFSIZE 256

F_NORETURN F_NONNULL
static void mainloop(const int fd, dnspacket_context_t* pctx, const bool use_cmsg) {
    dmn_assert(pctx);

    anysin_t asin;
    struct iovec iov = {
        .iov_base = mmap(NULL, gconfig.outer_max_response, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0),
        .iov_len  = 0
    };
    struct msghdr msg_hdr;
    char cmsg_buf[CMSG_BUFSIZE];

    memset(&msg_hdr, 0, sizeof(struct msghdr));
    msg_hdr.msg_name       = &asin;
    msg_hdr.msg_namelen    = sizeof(asin);
    msg_hdr.msg_iov        = &iov;
    msg_hdr.msg_iovlen     = 1;
    msg_hdr.msg_control    = use_cmsg ? cmsg_buf : NULL;
    msg_hdr.msg_controllen = use_cmsg ? CMSG_BUFSIZE : 0;

    while(1) {
        iov.iov_len = DNS_RECV_SIZE;
        const int buf_in_len = recvmsg(fd, &msg_hdr, 0);
        if(likely(buf_in_len >= 0)) {
#if DNSCURVE
            iov.iov_len = dnscurve_filter(pctx, &asin, (void*)iov.iov_base, buf_in_len);
#else
            iov.iov_len = process_dns_query(pctx, &asin, (void*)iov.iov_base, buf_in_len);
#endif
            if(likely(iov.iov_len)) {
                const int sent = sendmsg(fd, &msg_hdr, 0);
                if(unlikely(sent < 0)) {
                    satom_inc(&pctx->stats->p.udp.sendfail);
                    log_err("UDP sendmsg() of %li bytes failed with retval %i for client %s: %s", (long)iov.iov_len, sent, logf_anysin(&asin), logf_errno());
                }
            }
        }
        else {
            log_err("UDP recvmsg() error: %s", logf_errno());
        }
    }
}

#if HAVE_DECL___NR_RECVMMSG

#include <sys/syscall.h>
struct _gdnsd_mmsghdr { struct msghdr msg_hdr; unsigned msg_len; };
#ifndef MSG_WAITFORONE
# define MSG_WAITFORONE 0x10000
#endif
static inline int _gdnsd_recvmmsg(int fd, struct _gdnsd_mmsghdr *mmsg, unsigned vlen, unsigned flags, struct timespec* timeout) {
    return syscall(__NR_recvmmsg, fd, mmsg, vlen, flags, timeout);
}

F_NORETURN F_NONNULL
static void mainloop_mmsg(const unsigned width, const int fd, dnspacket_context_t* pctx, const bool use_cmsg) {
    dmn_assert(pctx);

    uint8_t* buf[width];
    struct iovec iov[width][1];
    struct _gdnsd_mmsghdr dgrams[width];
    char cmsg_buf[width][CMSG_BUFSIZE];
    anysin_t asin[width];

    // Set up packet buffers
    for (unsigned i = 0; i < width; i++)
        iov[i][0].iov_base = buf[i]
            = mmap(NULL, gconfig.outer_max_response, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);

    // Set up msg_hdr's
    for (unsigned i = 0; i < width; i++) {
        dgrams[i].msg_hdr.msg_iov        = iov[i];
        dgrams[i].msg_hdr.msg_iovlen     = 1;
        dgrams[i].msg_hdr.msg_name       = &asin[i];
        dgrams[i].msg_hdr.msg_namelen    = sizeof(anysin_t);
        dgrams[i].msg_hdr.msg_control    = use_cmsg ? cmsg_buf[i] : NULL;
        dgrams[i].msg_hdr.msg_controllen = use_cmsg ? CMSG_BUFSIZE : 0;
        dgrams[i].msg_hdr.msg_flags      = 0;
    }

    while(1) {
        // Reset iov lens for receiving
        for(unsigned i = 0; i < width; i++)
            iov[i][0].iov_len = DNS_RECV_SIZE;

        const int pkts = _gdnsd_recvmmsg(fd, dgrams, width, MSG_WAITFORONE, NULL);
        dmn_assert(pkts <= (int)width);
        if(likely(pkts > 0)) {
            for(int i = 0; i < pkts; i++)
#if DNSCURVE
                iov[i][0].iov_len = dnscurve_filter(pctx, &asin[i], buf[i], dgrams[i].msg_len);
#else
                iov[i][0].iov_len = process_dns_query(pctx, &asin[i], buf[i], dgrams[i].msg_len);
#endif
            for(int i = 0; i < pkts; i++) {
                if(likely(iov[i][0].iov_len)) {
                    const int sent = sendmsg(fd, &dgrams[i].msg_hdr, 0);
                    if(unlikely(sent < 0)) {
                        satom_inc(&pctx->stats->p.udp.sendfail);
                        log_err("UDP sendmsg() of %li bytes failed with retval %i for client %s: %s", (long)iov[i][0].iov_len, sent, logf_anysin(&asin[i]), logf_errno());
                    }
                }
            }
        }
        else {
            log_err("UDP recvmmsg() error: %s", logf_errno());
        }
    }
}

// runtime check for Linux > 2.6.33, for MSG_WAITFORONE support
//  why isn't there a runtime syscall that gives LINUX_VERSION_INFO?
#include <sys/utsname.h>
static bool use_recvmmsg(void) {
    int maj, min, rel;
    struct utsname uts;
    if(!uname(&uts) && !strcmp("Linux", uts.sysname)
        && sscanf(uts.release, "%d.%d.%d", &maj, &min, &rel) == 3) {
        if(maj > 2 || (maj == 2 && (min > 6 || (min == 6 && rel > 33))))
            return 1;
    }
    return 0;
}

#endif

static void thread_cleanup(void* arg) {
    dnspacket_context_t* pctx V_UNUSED = (dnspacket_context_t*) arg;
#if DNSCURVE
    destroy_thread_curve_keys(pctx);
#endif
}

void* dnsio_udp_start(void* addrconf_asvoid) {
    dmn_assert(addrconf_asvoid);
    const dns_addr_t* addrconf = (const dns_addr_t*) addrconf_asvoid;

    dnspacket_context_t* pctx = dnspacket_context_new(addrconf->udp_threadnum, true);

    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    pthread_cleanup_push(thread_cleanup, pctx);

#if DNSCURVE
    dnscurve_iothread_setup(pctx, addrconf);
#endif

    if(addrconf->udp_need_late_bind) {
        const anysin_t* asin = &addrconf->addr;
        const bool isv6 = asin->sa.sa_family == AF_INET6 ? true : false;
        while(bind(addrconf->udp_sock, (const struct sockaddr*)asin, isv6 ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in))) {
            if(errno != EADDRNOTAVAIL) {
                log_err("Failed late bind() of UDP socket to %s: %s.  This listener thread is now shutting down.  Late bind attempts for this socket will no longer be attempted!", logf_anysin(asin), logf_errno());
                pthread_exit(NULL);
            }
            sleep(addrconf->late_bind_secs);
        }
        log_info("Late bind() of UDP socket to %s succeeded, serving requests now", logf_anysin(asin));
    }

#if HAVE_DECL___NR_RECVMMSG
    if(addrconf->udp_recv_width > 1 && use_recvmmsg()) {
        log_debug("recvmmsg() with a width of %u enabled for UDP socket %s",
            addrconf->udp_recv_width, logf_anysin(&addrconf->addr));
        mainloop_mmsg(addrconf->udp_recv_width, addrconf->udp_sock, pctx, is_any_addr(&addrconf->addr));
    }
    else
#endif
    {
        mainloop(addrconf->udp_sock, pctx, is_any_addr(&addrconf->addr));
    }

    pthread_cleanup_pop(1);
}
