#    Copyright © 2011 Brandon L Black <blblack@gmail.com>
#
#    This file is part of gdnsd.
#
#    gdnsd is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    gdnsd is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
#

package _GDT;

# This should provide the common test functionality,
#  such as executing gdnsd in a forked child running
#  a test-specific dataset, and methods for running
#  queries against it and validating the output against
#  a pre-defined specification

require 5.008001;
use strict;
use warnings;
use POSIX ':sys_wait_h';
use Scalar::Util qw/looks_like_number/;
use File::Spec ();
use File::Temp ();
use Net::DNS::Resolver ();
use Net::DNS ();
use LWP::UserAgent ();
use Test::More ();
use Config;

# Hack around a Net::DNS 0.65 bug, so that we can
#  merely require 0.63 or higher
if($Net::DNS::VERSION == '0.65') {
    eval "require Net::DNS::RR; require AAAA66;";
    die "Failed to load Net::DNS 0.65 hack: $@" if $@;
    $Net::DNS::RR::_LOADED{'Net::DNS::RR::AAAA'}++;
}

my %SIGS;
{
    my $i = 0;
    defined $Config{sig_name} || die "No sigs?";
    foreach my $name (split(' ', $Config{sig_name})) {
        $SIGS{$name} = $i++;
    }
}

$File::Temp::KEEP_ALL = $ENV{KEEP_TEST_FILES} ? 1 : 0;

our $FIXED_TEST_OUT = 0;
our $TEST_RUNNER = "";
if($ENV{TEST_RUNNER}) {
    $FIXED_TEST_OUT = 1;
    $TEST_RUNNER = $ENV{TEST_RUNNER};
}
    
our $TESTPORT_START = $ENV{TESTPORT_START};
die "Test port start specification is not a number"
    unless looks_like_number($TESTPORT_START);

our $DNS_PORT = $TESTPORT_START;
our $HTTP_PORT = $TESTPORT_START + 1;
our $EXTRA_PORT  = $TESTPORT_START + 2;

our $saved_pid;

# Skip V6 tests if perl doesn't have the modules for it,
#  or it doesn't work at runtime.
our $HAVE_V6 = 1;
eval "require Socket6; require IO::Socket::INET6;";
if($@) {
    warn "IPv6 tests disabled (missing Socket6 and/or IO::Socket::INET6)";
    $HAVE_V6=0;
}
else {
    my $test_sock = IO::Socket::INET6->new(LocalAddr => '::1', LocalPort => $DNS_PORT);
    if(!$test_sock) {
        warn "IPv6 tests disabled (Cannot bind to [::1]:$DNS_PORT: $@)";
        $HAVE_V6=0;
    }
}

our $GDNSD_BIN = $ENV{INSTALLCHECK_SBINDIR}
    ? "$ENV{INSTALLCHECK_SBINDIR}/gdnsd"
    : "$ENV{TOP_BUILDDIR}/gdnsd/gdnsd";

# During installcheck, the default hardcoded plugin path
#  should work correctly for finding the installed trivial plugins
our $TRIVIAL_PLUGIN_PATH = $ENV{INSTALLCHECK_SBINDIR}
    ? "/xxx_does_not_exist"
    : "$ENV{TOP_BUILDDIR}/plugins/trivial/.libs";

our $RAND_LOOPS = $ENV{GDNSD_RTEST_LOOPS} || 100;

die "Cannot run testsuite as root" if ! $>;

my $CSV_TEMPLATE = 
    "uptime\r\n"
    . "([0-9]+)\r\n"
    . "noerror,refused,nxdomain,notimp,badvers,dropped,v6\r\n"
    . "([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)\r\n"
    . "udp_reqs,udp_sendfail,udp_tc,udp_edns,udp_edns_big,udp_edns_tc\r\n"
    . "([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)\r\n"
    . "tcp_reqs,tcp_recvfail,tcp_recvsize,tcp_sendfail\r\n"
    . "([0-9]+),([0-9]+),([0-9]+),([0-9]+)\r\n"
    . "curve_strm,curve_strm_fail,curve_txt,curve_txt_fail,curve_cache_hit,curve_cache_miss\r\n"
    . "([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)\r\n";

my $_useragent;
sub _get_daemon_csv_stats {
    $_useragent ||= LWP::UserAgent->new(
        protocols_allowed => ['http'],
        requests_redirectable => [],
        max_size => 10240,
        timeout => 3,
    );
    my $response = $_useragent->get("http://127.0.0.1:${HTTP_PORT}/csv");
    if(!$response) {
        return "No response...";
    }
    elsif($response->code != 200) {
        return "Response code was not 200. Response dump:\n" . $response->as_string("\n");
    }
    return $response->content;
}

sub check_stats_inner {
    my ($class, %to_check) = @_;
    my $content = _get_daemon_csv_stats();
    if($content !~ m/^${CSV_TEMPLATE}/s) {
       die "Content does not match CSV_TEMPLATE, content is: " . $content;
    }
    my $csv_vals = {
        uptime          => $1,
        noerror         => $2,
        refused         => $3,
        nxdomain        => $4,
        notimp          => $5,
        badvers         => $6,
        dropped         => $7,
        v6              => $8,
        udp_reqs        => $9,
        udp_sendfail    => $10,
        udp_tc          => $11,
        udp_edns        => $12,
        udp_edns_big    => $13,
        udp_edns_tc     => $14,
        tcp_reqs        => $15,
        tcp_recvfail    => $16,
        tcp_recvsize    => $17,
        tcp_sendfail    => $18,
        curve_strm      => $19,
        curve_strm_fail => $20,
        curve_txt       => $21,
        curve_txt_fail  => $22,
        curve_cache_hit  => $23,
        curve_cache_miss => $24,
    };

    ## use Data::Dumper; warn Dumper($csv_vals);

    foreach my $checkit (keys %to_check) {
        if($csv_vals->{$checkit} != $to_check{$checkit}) {
            die "$checkit mismatch, wanted " . $to_check{$checkit} . ", got " . $csv_vals->{$checkit};
        }
    }

    return;
}

sub check_stats {
    my ($class, %to_check) = @_;
    my $err;
    my $attempts = 0;
    while(1) {
        eval { $class->check_stats_inner(%to_check) };
        $err = $@;
        return unless $err;
        if($attempts++ < 10) {
            select(undef, undef, undef, 0.1 * $attempts);
        }
        else {
            die "Stats check failed repeatedly: $err";
        }
    }
}

sub spawn_daemon {
    my ($class, $cfgfile) = @_;

    my (undef, $cfdir, undef) = File::Spec->splitpath($cfgfile);
    $cfdir =~ s/\/$//;

    my $tmpdir;
    if(!$FIXED_TEST_OUT) {
        $tmpdir = File::Temp::tempdir(CLEANUP => 1);
        warn "TEMPDIR being kept is " . $tmpdir if $ENV{KEEP_TEST_FILES};
    }
    else {
        use File::Path ();
        $tmpdir = "./test_out/$0/";
        File::Path::make_path($tmpdir);
    }

    my $daemon_out = $tmpdir . '/gdnsd.out';
    my $cfgout = $tmpdir . '/gdnsd.conf';

    open(my $orig_fh, '<', $cfgfile)
        or die "Cannot open test configfile '$cfgfile' for reading: $!";
    open(my $out_fh, '>', $cfgout)
        or die "Cannot open test config text output '$cfgout' for writing: $!";

    my $dns_lspec = $HAVE_V6
        ? qq{[ 127.0.0.1, ::1 ]}
        : qq{127.0.0.1};

    my $http_lspec = $HAVE_V6
        ? qq{[ 127.0.0.1, ::1 ]}
        : qq{127.0.0.1};

    while(<$orig_fh>) {
        s/\@dns_lspec\@/$dns_lspec/g;
        s/\@http_lspec\@/$http_lspec/g;
        s/\@dns_port\@/$DNS_PORT/g;
        s/\@http_port\@/$HTTP_PORT/g;
        s/\@extra_port\@/$EXTRA_PORT/g;
        s/\@cfdir\@/$cfdir/g;
        s/\@trivpath\@/$TRIVIAL_PLUGIN_PATH/g;
        print $out_fh $_;
    }
    close($orig_fh) or die "Cannot close test configfile '$cfgfile': $!";
    close($out_fh) or die "Cannot close test config text output file '$cfgout': $!";

    my $exec_line = $TEST_RUNNER
        ? qq{$TEST_RUNNER $GDNSD_BIN -c $cfgout startfg}
        : qq{$GDNSD_BIN -c $cfgout startfg};

    my $pid = fork();
    die "Fork failed!" if !defined $pid;
    if(!$pid) { # child, exec daemon
        open(STDOUT, '>', $daemon_out)
            or die "Cannot open '$daemon_out' for writing as STDOUT: $!";
        open(STDIN, '<', '/dev/null')
            or die "Cannot open /dev/null for reading as STDIN: $!";
        open(STDERR, '>&STDOUT')
            or die "Cannot dup STDOUT to STDERR: $!";
        exec($exec_line);
    }

    # Used to kill -9 in END block
    $saved_pid = $pid;

    # With a test runner, we recheck for startup success every second for 300 seconds
    # Without a test runner, every 0.1 seconds for 10 seconds
    my $retry_delay = $TEST_RUNNER ? 1 : 0.1;
    my $retry = $TEST_RUNNER ? 300 : 100;

    select(undef, undef, undef, $retry_delay);
    while($retry--) {
        if(-f $daemon_out) {
            open(my $gdout_fh, '<', $daemon_out)
                or die "Cannot open '$daemon_out' for reading: $!";
            my $is_listening;
            while(<$gdout_fh>) {
                $is_listening = 1 if /\Qinfo: DNS listeners started\E$/;
            }
            close($gdout_fh)
                or die "Cannot close '$daemon_out': $!";
            if($is_listening) {
                kill($SIGS{USR1}, $pid); # enable logging of packet errors
                return $pid;
            }
        }
        select(undef, undef, undef, $retry_delay);
    }

    my $gdout = '';
    open(my $gdout_fh, '<', $daemon_out)
        or die "gdnsd failed to finish starting properly, and no output file could be found";
    while(<$gdout_fh>) { $gdout .= $_ }
    close($gdout_fh);
    die "gdnsd failed to finish starting properly.  output (if any):\n" . $gdout;
}

sub test_spawn_daemon {
    my ($class, $config) = @_;
    local $Test::Builder::Level = $Test::Builder::Level + 1;
    my $pid = eval{_GDT->spawn_daemon($config)};
    unless(Test::More::ok(!$@ && $pid)) {
        Test::More::diag("Cannot spawn daemon: $@");
        Test::More::BAIL_OUT($@);
    }

    return $pid;
}

my $_resolver;
my $_resolver6;

sub get_resolver {
    return $_resolver ||= Net::DNS::Resolver->new(
        recurse => 0,
        nameservers => [ '127.0.0.1'],
        port => $DNS_PORT,
        udp_timeout => 3,
        tcp_timeout => 3,
        retrans => 1,
        retry => 1,
    );
}

sub get_resolver6 {
    return $_resolver6 ||= Net::DNS::Resolver->new(
        recurse => 0,
        nameservers => [ '::1'],
        port => $DNS_PORT,
        udp_timeout => 3,
        tcp_timeout => 3,
        retrans => 1,
        retry => 1,
    );
}

# Creates a new Net::DNS::Packet which is a query response,
#  to compare with the real server response for correctness.
#  Args are: { headerparam => value }, $question, [ answers ], [ auths ], [ addtl ]
#  headers bits default to AA and QR on, the rest off, rcode NOERROR, opcode QUERY,
#  note this wont set ID properly, which is almost surely necessary, and should
#  match the question being asked.
sub mkanswer {
    my ($class, $headers, $question, $answers, $auths, $addtl) = @_;

    my $p = Net::DNS::Packet->new();

    # Add the user-supplied data
    $p->push('question', $question) if $question;
    $p->push('answer', $_) foreach @$answers;
    $p->push('authority', $_) foreach @$auths;
    $p->push('additional', $_) foreach @$addtl;

    # Enforce our defaults over Net::DNS's
    $p->header->qr(1);
    $p->header->aa(1);
    $p->header->tc(0);
    $p->header->rd(0);
    $p->header->cd(0);
    $p->header->ra(0);
    $p->header->ad(0);
    $p->header->opcode('QUERY');
    $p->header->rcode('NOERROR');
    $p->header->qdcount(0) if !$question;

    # Apply nondefault header settings from $headers
    foreach my $hfield (keys %$headers) {
        $p->header->$hfield($headers->{$hfield});
    }

    return $p;
}

sub compare_packets {
    my ($class, $answer, $compare, $limit_v4, $limit_v6) = @_;

    # Defined-ness checks
    if(!defined $answer) {
        if(!defined $compare) { return; }
        die "Answer was undefined when it shouldn't be";
    }
    elsif(!defined $compare) {
        die "Answer was defined when it shouldn't be";
    }

    my $result = _compare_contents($answer, $compare, $limit_v4, $limit_v6);
    if($result) {
        my $ans_str = $answer->string;
        my $comp_str = $compare->string;
        die "Packet mismatch! $result\n---Received:\n$ans_str\n---Wanted:\n$comp_str\n(Addr Limits: V4: $limit_v4 V6: $limit_v6)\n";
    }

    return $answer->answersize;
}

sub _compare_contents {
    my ($answer, $compare, $limit_v4, $limit_v6) = @_;

    # Header checks
    my $ahead = $answer->header;
    my $chead = $compare->header;
    return "ID mismatch" if $ahead->id != $chead->id;
    return "QR mismatch" if $ahead->qr != $chead->qr;
    return "OPCODE mismatch" if $ahead->opcode ne $chead->opcode;
    return "AA mismatch" if $ahead->aa != $chead->aa;
    return "TC mismatch" if $ahead->tc != $chead->tc;
    return "RD mismatch" if $ahead->rd != $chead->rd;
    return "CD mismatch" if $ahead->cd != $chead->cd;
    return "RA mismatch" if $ahead->ra != $chead->ra;
    return "AD mismatch" if $ahead->ad != $chead->ad;
    return "RCODE mismatch" if $ahead->rcode ne $chead->rcode;
    return "QDCOUNT mismatch" if $ahead->qdcount != $chead->qdcount;
    # These are skipped now because:
    # (a) They should be redundant to the more complex checks below, and
    # (b) They get really screwed up in the presence of $limit_v[46],
    #     which is really a pita to fix properly.
    #return "ANCOUNT mismatch" if $ahead->ancount != $chead->ancount;
    #return "NSCOUNT mismatch" if $ahead->nscount != $chead->nscount;
    #return "ADCOUNT mismatch" if $ahead->adcount != $chead->adcount;

    # Question checks
    my ($aquest) = $answer->question;
    my ($cquest) = $compare->question;
    my $aquest_str = $aquest ? $aquest->string : "";
    my $cquest_str = $cquest ? $cquest->string : "";
    return "Question mismatch" if $aquest_str ne $cquest_str;

    # Section content checks...
    my $ans_rv = _compare_sections("answer", $answer, $compare, $limit_v4, $limit_v6);
    return $ans_rv if $ans_rv;
    my $auth_rv = _compare_sections("authority", $answer, $compare, $limit_v4, $limit_v6);
    return $auth_rv if $auth_rv;
    my $addtl_rv = _compare_sections("additional", $answer, $compare, $limit_v4, $limit_v6);
    return $addtl_rv if $addtl_rv;

    return undef;
}

# Compare two sections by rrset, allowing random re-ordering of whole rrsets
#   as well as rotational offset within an rrset
sub _compare_sections {
    my ($section, $answer, $compare, $limit_v4, $limit_v6) = @_;
    my $a_rrsets = _get_rrsets($section, $answer);
    my $c_rrsets = _get_rrsets($section, $compare);

    return "$section mismatch: number of rrsets" unless @$a_rrsets == @$c_rrsets;

    my $i = 0;
    my $c_rrsets_href = { map { $i++ => $_ } @$c_rrsets };

    for(my $x = 0; $x < @$a_rrsets; $x++) {
        my $rrtype = $a_rrsets->[$x]->[0]->type;
        my $name = $a_rrsets->[$x]->[0]->name;
        my $match_found;
        foreach my $k (keys %$c_rrsets_href) {
            my $crrtype = $c_rrsets_href->{$k}->[0]->type;
            my $cname = $c_rrsets_href->{$k}->[0]->name;
            next if $crrtype ne $rrtype || $cname ne $name;
            my $cmp_err = _compare_rrsets($a_rrsets->[$x], $c_rrsets_href->{$k}, $limit_v4, $limit_v6);
            if($cmp_err) {
                return "$section mismatch: response rrset $name $rrtype: $cmp_err";
            }
            else {
                $match_found = 1;
                delete $c_rrsets_href->{$k};
                last;
            }
        }
        return "$section mismatch: response rrset $name $rrtype not matched" unless $match_found;
    }

    return undef;
}

# Compare two rrsets, allowing for a rotational offset and A/AAAA size limits
sub _compare_rrsets {
    my ($a_rrset, $c_rrset, $limit_v4, $limit_v6) = @_;

    my $rrtype = $c_rrset->[0]->type;
    my $limit = $rrtype eq 'A'
        ? $limit_v4
        : $rrtype eq 'AAAA'
            ? $limit_v6
            : @$c_rrset;

    if($limit > @$c_rrset || $limit == 0) { $limit = @$c_rrset; }

    # At this point, regardless of rrtype or the presence
    # of limit arguments/config, we know we need $a_rrset
    # to match some in-order $limit subset of $c_rrset,
    # allowing for wraparound

    return "rrset size mismatch" if @$a_rrset != $limit;

    for(my $i = 0; $i < @$c_rrset; $i++) {
        if($a_rrset->[0]->string eq $c_rrset->[$i]->string) {
           for(my $j = 1; $j < @$a_rrset; $j++) {
               return "specific rr mismatch ($j in response)" if $a_rrset->[$j]->string ne $c_rrset->[($i + $j) % @$c_rrset]->string;
           }
           return undef; # success
        }
    }
    return "no match sequence start point";
}

# Split the RRs of a section (of a packet) into rrsets
sub _get_rrsets {
    my ($section, $packet) = @_;

    my $rrsets = [];
    my $rrsets_idx = 0;

    my @rrs = $packet->$section;
    for(my $rrs_idx = 0; $rrs_idx < @rrs; $rrs_idx++) {
        push(@{$rrsets->[$rrsets_idx]}, $rrs[$rrs_idx]);
        # If 1+ RRs remain after the one just pushed and the
        #  next has a different name or type than the current,
        #  increment the rrsets_idx.
        $rrsets_idx++ if $rrs_idx < @rrs - 1 &&
            ( $rrs[$rrs_idx]->type ne $rrs[$rrs_idx + 1]->type
            || $rrs[$rrs_idx]->name ne $rrs[$rrs_idx + 1]->name )
    }

    return $rrsets;
}

sub query_server {
    my ($class, $query, $expected, $res, $ro, $limit_v4, $limit_v6) = @_;

    $limit_v4 ||= 0;
    $limit_v6 ||= 0;

    my %saveopts;
    foreach my $k (keys %$ro) {
        $saveopts{$k} = $res->$k();
        $res->$k($ro->{$k})
    }

    my $size = 0;

    if($expected) {
        $expected->header->id($query->header->id);
        $size = _GDT->compare_packets($res->send($query), $expected, $limit_v4, $limit_v6);
    }
    else {
        $res->bgsend($query);
    }

    # restore altered resolver options
    foreach my $k (keys %saveopts) {
       $res->$k($saveopts{$k});
    }

    return $size;
}

my %stats_accum = (
    noerror      => 0,
    refused      => 0,
    nxdomain     => 0,
    notimp       => 0,
    badvers      => 0,
    dropped      => 0,
    v6           => 0,
    udp_reqs     => 0,
    udp_sendfail => 0,
    udp_tc       => 0,
    udp_edns     => 0,
    udp_edns_big => 0,
    udp_edns_tc  => 0,
    tcp_reqs     => 0,
    tcp_recvfail => 0,
    tcp_recvsize => 0,
    tcp_sendfail => 0,
);

sub test_dns {
    my ($class, %args) = @_;
    local $Test::Builder::Level = $Test::Builder::Level + 1;

    $args{qname}    ||= '.';
    $args{qtype}    ||= 'A';
    $args{header}   ||= {};
    $args{stats}    ||= [qw/udp_reqs noerror/];
    $args{resopts}  ||= {};
    $args{limit_v4} ||= 0;
    $args{limit_v6} ||= 0;

    foreach my $sec (qw/answer auth addtl/) {
        my $aref;
        if(!exists $args{$sec}) {
            $aref = $args{$sec} = [];
        }
        elsif(!ref $args{$sec} || ref $args{$sec} ne 'ARRAY') {
            $aref = $args{$sec} = [$args{$sec}];
        }
        else {
            $aref = $args{$sec};
        }
        map { if(!ref $aref->[$_]) { $aref->[$_] = Net::DNS::RR->new_from_string($aref->[$_]) } } (0..$#$aref);
    }

    my $question = $args{qpacket} || Net::DNS::Packet->new($args{qname}, $args{qtype});
    my $answer = $args{nores}
        ? undef
        : $class->mkanswer(
            $args{header},
            $args{noresq} ? undef : $question->question(),
            $args{answer},
            $args{auth},
            $args{addtl},
        );

    my ($size4, $size6) = (0, 0);

    if(!$args{v6_only}) {
        foreach my $stat (@{$args{stats}}) {
            $stats_accum{$stat}++;
        }
        my $res4 = $class->get_resolver();
        $size4 = eval { $class->query_server($question, $answer, $res4, $args{resopts}, $args{limit_v4}, $args{limit_v6}) };
        if($@) {
            Test::More::ok(0);
            Test::More::diag("IPv4 query failed: $@");
            return $size4;
        }
    }

    if(!$args{v4_only} && $HAVE_V6) {
        foreach my $stat (@{$args{stats}}) {
            $stats_accum{$stat}++;
            if($stat eq 'tcp_reqs' || $stat eq 'udp_reqs') {
                $stats_accum{v6}++;
            }
        }
        my $res6 = $class->get_resolver6();
        $size6 = eval { $class->query_server($question, $answer, $res6, $args{resopts}, $args{limit_v4}, $args{limit_v6}) };
        if($@) {
            Test::More::ok(0);
            Test::More::diag("IPv6 query failed: $@");
            return $size6;
        }
    }

    if($HAVE_V6 && !$args{v4_only} && !$args{v6_only} && $size4 != $size6) {
        Test::More::ok(0);
        Test::More::diag("Testsuite bug? IPv4/IPv6 size mismatch");
        return 0;
    }

    eval { $class->check_stats(%stats_accum) };
    Test::More::ok(!$@) or Test::More::diag("Stats check: $@");

    return $size4 || $size6;
}

sub test_stats {
    my $class = shift;
    local $Test::Builder::Level = $Test::Builder::Level + 1;
    eval { $class->check_stats(%stats_accum) };
    Test::More::ok(!$@) or Test::More::diag("Stats check: $@");
}

sub stats_inc { shift; $stats_accum{$_}++ foreach (@_); }

sub test_kill_daemon {
    my ($class, $pid) = @_;
    local $Test::Builder::Level = $Test::Builder::Level + 1;

    if(!$pid) {
        Test::More::ok(0);
        Test::More::BAIL_OUT("Test Bug: no pid specified?");
    }

    if(!kill(0, $pid)) {
        Test::More::ok(0);
        Test::More::BAIL_OUT("Daemon at pid $pid was dead before we tried to shut it down");
    }
    else {
        eval {
            local $SIG{ALRM} = sub { die "Failed to kill daemon cleanly at pid $pid"; };
            alarm(5);
            kill(2, $pid);
            waitpid($pid, 0);
        };
        if($@) {
            Test::More::ok(0);
            Test::More::BAIL_OUT($@);
        }
    }

    Test::More::ok(1);
}

END { kill(9, $saved_pid) if $saved_pid; }
1;
